/**
 * @implements API.Service
 */
export class Web3Storage implements API.Service {
    /**
     * @hidden
     * @param {string} token
     */
    static headers(token: string): {
        Authorization: string;
        'X-Client': string;
    };
    /**
     * @param {API.Service} service
     * @param {Iterable<API.Filelike>} files
     * @param {API.PutOptions} [options]
     * @returns {Promise<API.CIDString>}
     */
    static put({ endpoint, token }: API.Service, files: Iterable<API.Filelike>, { onRootCidReady, onStoredChunk, maxRetries, name }?: API.PutOptions | undefined): Promise<API.CIDString>;
    /**
     * @param {API.Service} service
     * @param {string} cid
     * @returns {Promise<import('./lib/interface.js').Web3Response | null>}
     */
    static get({ endpoint, token }: API.Service, cid: string): Promise<import('./lib/interface.js').Web3Response | null>;
    /**
     * @param {API.Service} service
     * @param {string} cid
     * @returns {Promise<API.CIDString>}
     */
    static delete({ endpoint, token }: API.Service, cid: string): Promise<API.CIDString>;
    /**
     * Constructs a client bound to the given `options.token` and
     * `options.endpoint`.
     *
     * @example
     * ```js
     * import { Web3Storage, Blob } from "web3.storage"
     * const client = new Web3Storage({ token: API_TOKEN })
     * const { car, rootCid } = await client.pack(new Blob(['hello world']))
     * const cid = await client.store(car)
     * console.assert(cid === rootCid, 'The service should store the files with the `rootCid` I created')
     * ```
     * Optionally you could pass an alternative API endpoint (e.g. for testing)
     * @example
     * ```js
     * import { Web3Storage } from "web3.storage"
     * const client = new Web3Storage({
     *   token: API_TOKEN
     *   endpoint: new URL('http://localhost:8080/')
     * })
     * ```
     *
     * @param {{token: string, endpoint?:URL}} options
     */
    constructor({ token, endpoint }: {
        token: string;
        endpoint?: URL;
    });
    /**
     * Authorization token.
     *
     * @readonly
     */
    readonly token: string;
    /**
     * Service API endpoint `URL`.
     * @readonly
     */
    readonly endpoint: URL;
    /**
     * Uploads files to web3.storage. Files are hashed in the client and uploaded as a single
     * [Content Addressed Archive(CAR)](https://github.com/ipld/specs/blob/master/block-layer/content-addressable-archives.md).
     * Takes a [Blob](https://developer.mozilla.org/en-US/docs/Web/API/Blob/Blob)
     *
     * Returns the corresponding Content Identifier (CID).
     *
     * @example
     * ```js
     * const data = 'Hello world'
     * const {root, car} = client.pack('Hello world')
     * const cid = await client.store(car)
     * console.assert(cid === root)
     * ```
     * @param {Iterable<API.Filelike>} files
     * @param {API.PutOptions} [options]
     */
    put(files: Iterable<API.Filelike>, options?: API.PutOptions | undefined): Promise<API.CIDString>;
    /**
     * Fetch the Content Addressed Archive by it's root CID
     * @param {string} cid
     */
    get(cid: string): Promise<API.Web3Response | null>;
    /**
     *
     * @param {string} cid
     */
    delete(cid: string): Promise<API.CIDString>;
}
import { File } from "./platform.js";
import { Blob } from "./platform.js";
import * as API from "./lib/interface.js";
export { File, Blob };
//# sourceMappingURL=lib.d.ts.map